#!/bin/sh
# $Id: pcScan.tcl,v 1.56 2002/05/05 09:36:23 gdmr Exp $

# Copyright (c) 2000 Division of Informatics, University of Edinburgh
# 
# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation, and that the name of the University of Edinburgh not be used
# in advertising or publicity pertaining to distribution of the software
# without specific, written prior permission.  The University of Edinburgh
# makes no representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied warranty.
#
# Comments are invited.  Please send to gdmr@dcs.ed.ac.uk.

# Scan HP ProCurve switches, comparing what SNMP claims is
# configured against what's in a "definitive" configuration file.

# tclsh is in a different place on Solaris and Linux, and we can't
# assume that it's in our PATH as we may be running from a cron job.
# This is the standard tcl hack, only with more than one shell line...
#\
case "`uname -sr`" in
#\
	'SunOS 5'*)	TCLSH=/usr/local/bin/tclsh ;;
#\
	Linux*)		TCLSH=/usr/tcl8/bin/tclsh ;;
#\
	 *)		echo "Don't know where to find tclsh" ; exit ;;
#\
esac
#\
exec $TCLSH "$0" "$@"
# It's tcl from here on...

# Auto-load from the directory that this script lives in
lappend auto_path [file dirname $argv0]

#
# Get the switch configuration name from the command line
# and any flags
#
proc Usage {} {
   global argv0
   puts stderr "$argv0 \[-update] \[-showDown] \[-Community] config-file"
}

# Fault-finder cutoff time
set ffCutoff [expr [clock seconds] - 86400 * 5]

set doShowDown 0
set doUpdate 0
set fwdLimit 0
set verbose 1
set debug 0
foreach a $argv {
   switch -glob -- $a {
      -update	{ set doUpdate 1 }
      -down     -
      -showDown -
      -showdown { set doShowDown 1 }
      -hideDown -
      -hidedown { set doShowDown -1 }
      -quiet    -
      -q        { set verbose 0 }
      -verbose  -
      -v        { incr verbose }
      -debug    { set debug 1 }
      -C*       { regexp {^-C(.*)$} $a all community }
      -F*       { if {![regexp {^-F([0-9]+)$} $a all fwdLimit]} {
                     Usage ; exit 1 }}
      -L*       { regexp {^-L(.*)$} $a all lockOpt }
      -*	{ Usage ; exit 1 }
      *		{ if [info exists configFile] { Usage ; exit 1 }
		  set configFile $a }
   }
}

if {$debug} {puts "ffCutoff: [clock format $ffCutoff]"}

if {![info exists configFile]} {
   Usage
   exit 1
}

# What's legal in the configuration file?
set confCommands [parseCommands]

# Read the configuration file
proc readConfig {configFile confCommands} {
   global debug
   puts "Reading configuration file \"$configFile\"..."
   if [catch {open $configFile r} cf] {
      puts stderr "Read switch configuration: $cf"
      exit 2
   } else {
      if [catch {split [read $cf] "\n"} confFile] {
         puts stderr "Error reading $configFile: $confFile"
         exit 2
      }
      close $cf
   }
   # Parse it if it read in OK
   set bogusCommand 0
   foreach confLine $confFile {
      set confLine [string trim $confLine]
      if [string length $confLine] {
         set cfw [lindex [split $confLine] 0]
         if [string compare [string index $cfw 0] "#"] {
            # Not a comment, so...
            if {![string compare $cfw "portfile"] || \
                  ![string compare $cfw "shadowportfile"]} {
               # Read a sub-file.  "shadow..." gets loaded here
               # but ignored by other code that reads the configs.
               # ("eval" loses any quotes there might be)
               set pfn [lindex [split $confLine] 1]
               eval readConfig $pfn [list "port"]
            } elseif {![string compare $cfw "vlanfile"]} {
               # Read a sub-file.
               # ("eval" loses any quotes there might be)
               set pfn [lindex [split $confLine] 1]
               eval readConfig $pfn [list "vlan"]
            } elseif {[lsearch -exact $confCommands $cfw] >= 0} {
               # Validated against legal commands
               if [catch {eval $confLine} problem] {
                  puts stderr "Error in \"$confLine\""
                  if {$debug} {puts stderr "$problem"}
                  set bogusCommand 1
               }
            } else {
               # Something we don't recognise
               puts stderr "Unknown/disallowed configuration option \"$cfw\""
               set bogusCommand 1
            }
         }
      }
   }
   # Did it parse?
   if {$bogusCommand || ![parseOK]} {
      puts stderr "Terminating due to errors in $configFile"
      exit 2
   }
}

readConfig $configFile $confCommands

# Initialisation...
set genericUpdates {}
set portUpdates {}
set vlanDeletes {}
set trunkingInUse 0

set configured [getPortList]
#puts "Configured: $configured"
set skipList [getSkipList]
if {[llength $skipList] > 0} {
   puts "Skip: $skipList"
}
set lockList [getLockList]
if [info exists lockOpt] {
   foreach l [split $lockOpt ","] {
      lappend lockList $l
   }
}
if {[llength $lockList] > 0} {
   puts "Locked: $lockList"
}

if {![info exists community]} {
   set community [getCommunity]
}
snmpInit [getTarget] $community

if [catch {snmpStpAdminStatus} stpState] {
   puts stderr "** No response from [getTarget]: $stpState"
   exit 98
}
if {$stpState} {
   set stpInfo [snmpStpInfo]
   set stpPriority [lindex $stpInfo 0]
   set stpRootCost [lindex $stpInfo 1]
   set stpRootPort [lindex $stpInfo 2]
   set stpDesRoot  [lindex $stpInfo 3]
} else {
   set stpPriority -1
   set stpRootCost 0
   set stpRootPort -1
   set stpDesRoot "*unknown*"
}
#puts "STP: prio $stpPriority, rCost $stpRootCost, rPort $stpRootPort, dRoot $stpDesRoot"

set bootTime [expr [clock seconds] - [snmpUpTime]]

if {$verbose} {
   puts "Switch [getTarget] ([snmpLocation])"
   set descr [snmpDescr]
   if [regexp "(.*) \\(" $descr all descr] {}
   puts "\t$descr"
   set httpStuff [snmpHpHttpStuff]
   set health [lindex $httpStuff 0]
   if [string compare $health "ok"] {set health "** $health **" }
   set sno [string trim [lindex $httpStuff 1]]
   set ano [string trim [lindex $httpStuff 2]]
   puts "\thealth \"$health\", s/n $sno"
   set since "[clock format $bootTime]"
   puts "\tup since $since"
   if {$stpState} {
      if {$stpRootPort == 0 && $stpRootCost == 0} {
         set weAreRoot ", **ROOT**"
      } else {
         set weAreRoot ""
      }
      puts [format "\tSTP enabled, priority %#04x%s" $stpPriority $weAreRoot]
   } else {
      set stpWanted [getStpWanted]
      if {$stpWanted} {
         puts "\t** Spanning tree protocol disabled"
      } else {
         if {$debug} {puts "\tSPT disabled"}
      }
   }
   set tab "\t"
} else {
   set tab ""
}
foreach cs [snmpHpChassisSensors] {
   set descr [lindex $cs 0]
   set state [lindex $cs 1]
   set warnings [lindex $cs 2]
   set failures [lindex $cs 3]
   if {$verbose > 1 || 
         $warnings > 0 || $failures > 0 || \
         ([string compare $state "good"] && \
            [string compare $state "notPresent"])} {
      if {$warnings > 0} {
         if {$failures > 0} {
            set wf " (warnings: $warnings, failures: $failures)"
         } else {
            set wf " (warnings: $warnings)"
         }
      } else {
         set wf " (failures: $failures)"
      }
      puts "$tab$descr: $state$wf"
   }
}

puts "Grabbing fault-finder info..."
set ffInfo [snmpFfInfo]
foreach f $ffInfo {
   set ffStamp [lindex $f 0]
   set ffPhys [lindex $f 1]
   set ffType [lindex $f 2]
   set ffNew [lindex $f 3]
   if {! [string compare $ffNew "new"]} {
      #if {$debug} {puts "New ff: $f"}
      if [info exists ffPortInfo($ffPhys)] {
         lappend ffPortInfo($ffPhys) [list $ffStamp $ffType]
      } else {
         set ffPortInfo($ffPhys) [list [list $ffStamp $ffType]]
      }
   } else {
      if {$debug} {puts "Old ff: $f"}
   }
}

puts "Scanning for VLANs..."
set pcVlanList [snmpVlanScan]
set confVlanList [getVlanTagList]
puts "VLANs: checking configuration against switch..."
foreach v $pcVlanList {
   if {[lsearch -exact $confVlanList $v] < 0} {
      set n [snmpVlanTagToName $v]
      puts " - $n ($v) not in configuration"
      lappend vlanDeletes [list snmpDeleteVlan $n $v]
   }
}
foreach v $confVlanList {
   if {[lsearch -exact $pcVlanList $v] < 0} {
      set n [getVlanFullName $v]
      puts " - $n ($v) not on switch"
      lappend genericUpdates [list snmpCreateVlan $n $v]
   }
}
if {[llength $vlanDeletes] > 0} {
   puts "** Forcing -showDown due to unconfigured VLANs"
   set doShowDown 1
}

if {$fwdLimit > 0} {
   puts "Grabbing port forwarding table..."
   set pft [snmpPortFwdTable]
   foreach fe $pft {
      set i [lindex $fe 0]
      set m [lindex $fe 1]
      if [info exists pfl($i)] {
         lappend pfl($i) $m
      } else {
         set pfl($i) [list $m]
      }
   }
}
#foreach i [array names pfl] {
#   puts "$i -> $pfl($i)"
#}

puts "Scanning for interfaces..."
set actual [snmpEtherIfIndices]

proc putVlanList {label tags nameFn} {
   set text ""
   set wrap 1
   foreach t $tags {
      set n [$nameFn $t]
      if {[string length $text] > 0} {
         append text ", "
         if {[expr [incr wrap] & 3] == 0} {
            append text "\n\t\t"
         }
      }
      append text "$n ($t)"
   }
   puts "\t$label: $text"
}

proc fwdListText {pi} {
   global pfl
   global fwdLimit
   if [info exists pfl($pi)] {
      set n [llength $pfl($pi)]
      if {$n > $fwdLimit} {
         return "\t-> ($n addresses)"
      } else {
         set text ""
         set wrap 0
         foreach a $pfl($pi) {
            if {[string length $text] > 0} {
               append text ", "
               if {[expr [incr wrap] % 3] == 0} {
                  append text "\n\t   "
               }
            }
            append text "$a"
         }
         return "\t-> $text"
      }
   }
}

proc splitSpeed {spec speed duplex} {
   upvar $speed s
   upvar $duplex d
   regexp {^(10*).*(HD|FD)$} $spec all s d
}

set reel [snmpRmonEtherErrorLabels]
set rpsl [snmpRmonPortSizeLabels]
set elel [snmpPortEtherLikeErrorsLabels]
set elef [snmpPortEtherLikeErrorsDispflags]

proc processOne {pi pn portName} {
   global upDown
   global skipList
   global lockList
   global genericUpdates
   global vlanDeletes
   global portUpdates
   global verbose
   global doShowDown
   global fwdLimit
   global debug
   global stpState
   global ffPortInfo
   global ffCutoff
   global bootTime
   global reel
   global rpsl
   global elel
   global elef
   global stpRootPort
   global stpRootCost
   global trunkingInUse
   # Is it up or down?
   if [catch {set ifState [snmpIfUp $pi]} problem] {
      puts "$pn:\t$portName NOT ON SWITCH STACK?"
      if {$debug} {puts "\t$problem"}
      return
   }
   if {$ifState} {
      set upOrDown "up"
      set upDown($pi) 1
   } else {
      set upOrDown "DOWN"
      set upDown($pi) 0
   }
   # Should we ignore it completely?
   if {[lsearch -exact $skipList $pn] >= 0} return 
   # What's its current admin state?  If it's not as expected then
   # queue a command to fix it up.
   set adminState [snmpIfAdmin $pi]
   set hpAdminState [snmpHpPortAdmin $pi]
   set confState [getPortEnable $pn]
   #puts "$pn: conf $confState, admin $adminState, hp $hpAdminState"
   if {$adminState == 0 || $hpAdminState == 0} {
      # One of the adminStates was down or disabled.
      if {$adminState} {
         # Must have been the "HP" state that was disabled.
         # We never set that here, so it must have been because of
         # a bug or through another management interface...
         set upOrDown "DISABLED!!"
      } else {
         # (Probably) turned off through the usual mechanisms.
         set upOrDown "DISABLED"
      }
      # Either way, check to see if it should be the way it is.
      if {$confState} {
         set sbd "\n\tPort should be enabled"
         lappend portUpdates [list snmpIfSetAdmin $pi "up"]
      } else {
         set sbd ""
      }
   } else {
      # The port's enabled.  Should it be?
      if {!$confState} {
         set sbd "\n\tPort should be DISABLED"
         lappend portUpdates [list snmpIfSetAdmin $pi "down"]
      } else {
         set sbd ""
      }
   }
   # Get port type, speed, full/half-duplex
   set mauType [snmpMauType $pi]
   if {[regexp "MauType(.*)" $mauType all mt]} {
      set mauType $mt
   }
   set mauAuto [snmpMauAuto $pi]
   if {$mauAuto == "enabled"} {
      set as "ngtd"
   } else {
      set as "set"
   }
   # Spanning-tree state and mode, if applicable
   if {$stpState} {
      set stpPS [snmpStpPortState $pi]
      set stpPC [snmpStpPortCost $pi]
      set stpMode [snmpStpPortMode $pi]
      set stp ", $stpMode/$stpPC/$stpPS"
      if {$pi == $stpRootPort} {
         append stp "/$stpRootCost"
      }
   } else {
      set stp ""
   }
   # Print (or be ready to...)
   set portText "$pn:\t$portName $upOrDown, $as $mauType"
   # What's its default VLAN?  Or is it trunked?
   set tg [snmpIfTrunkGroup $pi]
   if {$tg > 0} {
      # It's trunked
      append portText "$stp, on trunk #$tg"
      set trunkingInUse 1
   } else {
      set vlanLists [snmpVlanLists $pi]
      set defVlan [lindex $vlanLists 0]
      append portText "$stp, on [snmpVlanTagToName $defVlan] ($defVlan)"
   }
   append portText $sbd
   #puts "$pn: $portName $ifState $doShowDown $debug $verbose"
   if {$ifState || $doShowDown >= 0 || $debug} {
      if {$ifState || $doShowDown >= 0} {
         puts "$portText"
      } else {
         puts "($portText)"
      }
      set printedText 1
   } else {
      set printedText 0
   }
   # Compare mauType/mauAuto against the configuration.
   #   should be negotiated + actually negotiated: OK
   #      "   "    "   "    + actually set: complain, queue correction
   #   should be set + actually negotiated: complain, queue correction
   #      "   "   "  + actually set: check speed, complain and queue
   #                                 correction if necessary
   set confSpeed [getPortSpeed $pn]
   if {![string compare $confSpeed "KEEP"]} {
      if {$debug} {puts "\tKeep existing speed"}
   } elseif {![string compare $confSpeed "-"]} {
      if {![string compare $mauAuto "enabled"]} {
         if {$debug} {puts "\tSpeed set and configured auto"}
      } else {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         puts "\tSpeed should be auto-negotiated"
         if {[lsearch -exact $lockList $pn] < 0} {
            lappend portUpdates [list snmpSetSpeed $pi $pn "-"]
         } else {
            puts stderr "\t** Port configuration is locked"
         }
      }
   } else {
      if {![string compare $mauAuto "enabled"]} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         puts "\tSpeed should be set to $confSpeed"
         if {[lsearch -exact $lockList $pn] < 0} {
            lappend portUpdates [list snmpSetSpeed $pi $pn $confSpeed]
         } else {
            puts stderr "\t** Port configuration is locked"
         }
      } else {
         splitSpeed $confSpeed cs ca
         splitSpeed $mauType ms ma
         if {$cs == $ms && $ca == $ma} {
            if {$debug} {puts "\tSpeed set $mauType and configured $confSpeed"}
         } else {
            if {!$printedText} {
               puts "$portText"
               set printedText 1
            }
            puts "\tSpeed should be set to $confSpeed"
            if {[lsearch -exact $lockList $pn] < 0} {
               lappend portUpdates [list snmpSetSpeed $pi $pn $confSpeed]
            } else {
               puts stderr "\t** Port configuration is locked"
            }
         }
      }
   }
   # STP mode
   if {$stpState} {
      set stpMConf [getStpMode $pn]
      set stpCConf [getStpCost $pn]
      if {$debug} {
         puts "\tSTP mode: conf $stpMConf, actual $stpMode"
         puts "\tSTP cost: conf $stpCConf, actual $stpPC"
      }
      if {[string compare $stpMConf $stpMode] || \
            ($stpCConf > 0 && $stpCConf != $stpPC)} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         puts "\tSTP should be \"$stpMConf/$stpCConf\""
         if {[lsearch -exact $lockList $pn] < 0} {
            if {$stpCConf > 0} {
               if {$debug} { puts "\tStp cost has changed" }
               lappend portUpdates \
                     [list snmpSetStpPort $pi $stpMConf $stpCConf]
               # XXX: because of a bug/feature/whatever in the firmware
               # we have to re-enable the port if we set its cost!  We
               # should really only do this if the port isn't configured
               # to be disabled!
               lappend portUpdates [list snmpIfSetAdmin $pi "up"]
            } else {
               if {$debug} { puts "\tStp cost unchanged" }
               lappend portUpdates \
                     [list snmpSetStpPort $pi $stpMConf]
            }
         } else {
            puts stderr "\t** Port configuration is locked"
         }
      }
   }
   # VLANs, if not trunked
   if {$tg == 0} {
      set updateVlans 0
      set confDefVlan [getPortDefVlan $pn]
      if {$defVlan != $confDefVlan} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         puts "\tShould be on [getVlanFullName $confDefVlan] ($confDefVlan)"
         set updateVlans 1
      }
      set pcTaggedVlanList [lindex $vlanLists 1]
      set confTaggedVlanList [getPortVlans $pn]
      set isList {}
      set alsoList {}
      set shouldList {}
      foreach t $pcTaggedVlanList {
         if {[lsearch -exact $confTaggedVlanList $t] >= 0} {
            lappend isList $t
         } else {
            lappend alsoList $t
            set updateVlans 1
         }
      }
      foreach t $confTaggedVlanList {
         if {[lsearch -exact $pcTaggedVlanList $t] < 0} {
            lappend shouldList $t
            set updateVlans 1
         }
      }
      if {$verbose && [llength $isList] > 0} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         putVlanList "Tagged VLANs" $isList snmpVlanTagToName
      }
      if {[llength $alsoList] > 0} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         putVlanList "Extra tagged VLANs" $alsoList snmpVlanTagToName
      }
      if {[llength $shouldList] > 0} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         putVlanList "Missing tagged VLANs" $shouldList getVlanFullName
      }
      # Do the VLAN update if either the port isn't locked
      # OR a new VLAN has been created or deleted.
      if {$updateVlans} {
         #puts "LL: $lockList, GU: $genericUpdates"
         if {[lsearch -exact $lockList $pn] < 0 || \
               [llength $genericUpdates] > 0 || \
               [llength $vlanDeletes] > 0} {
            lappend portUpdates [list snmpSetVlans \
               $pi $pn $confDefVlan $confTaggedVlanList]
         } else {
            puts stderr "\t** VLANs: port configuration is locked"
         }
      }
   }
   # Fault-finder, if any
   if [info exists ffPortInfo($pi)] {
      foreach f $ffPortInfo($pi) {
         # Divide by 100 by chopping the last two digits off the end of
         # the timeticks value...
         set fftt [lindex $f 0]
         set fft [string range $fftt 0 [expr [string length $fftt] - 3]]
         set ffStamp [expr $bootTime + $fft]
         if {$ffStamp > $ffCutoff} {
            if {!$printedText} {
               puts "$portText"
               set printedText 1
            }
            puts "\t** Fault-finder: [lindex $f 1] at [clock format $ffStamp]"
         } elseif {$debug} {
            puts "\t(** Fault-finder: [lindex $f 1] at [clock format $ffStamp])"
         }
      }
   }
   # Port statistics
   if {$verbose && ($ifState || $printedText || $doShowDown >= 0)} {
      set stats [snmpIfStats $pi]
      set st "\tFrames in: [lindex $stats 1]"
      set ie [lindex $stats 2]
      if {$ie != 0} {
         append st " ($ie error"
         if {$ie != 1} {append st "s"}
         append st ")"
      }
      append st ", out: [lindex $stats 4]"
      set oe [lindex $stats 5]
      if {$oe != 0} {
         append st " ($oe error"
         if {$oe != 1} {append st "s"}
         append st ")"
      }
      set ase [snmpMauAvailableExits $pi]
      if {$ase > 0} {
         append st "; link down: $ase"
      }
      if {!$printedText} {
         puts "$portText"
         set printedText 1
      }
      puts "$st"
      # Etherlike errors
      set ele [snmpPortEtherLikeErrors $pi]
      set eleshow 0
      foreach ee $ele el $elel ef $elef {
         if {$ee > 0} {
            if {$ef > 0} {set eleshow 1}
            if {$verbose > 1 || $ef >= 0} {
               if [info exists elet] {
                  append elet ", $ee $el"
               } else {
                  set elet "$ee $el"
               }
            }
         }
      }
      if {$eleshow} {
         puts "\t-- Ether errors: $elet"
      }
      # Input errors
      if {$ie > 0} {
         set ree [snmpRmonEtherErrors $pi]
         #puts "ree $pi: $ree"
         #puts "reel: $reel"
         foreach ee $ree el $reel {
            if {$ee > 0} {
               if [info exists et] {
                  append et ", $ee $el"
               } else {
                  set et "$ee $el"
               }
            }
         }
         if [info exists et] {
            puts "\t-- Input errors: $et"
         }
      }
      # Packet size counters
      if {$verbose > 1} {
         set rpsc [snmpRmonPortSizeCounters $pi]
         puts "\t[lindex $rpsl 0]: [lindex $rpsc 0], [lindex $rpsl 1]: [lindex $rpsc 1], [lindex $rpsl 2]: [lindex $rpsc 2]"
         puts "\t[lindex $rpsl 3]: [lindex $rpsc 3], [lindex $rpsl 4]: [lindex $rpsc 4], [lindex $rpsl 5]: [lindex $rpsc 5]"
      }
   }
   # Port forwarding table
   if {$verbose && $fwdLimit > 0} {
      set flt [fwdListText $pi]
      if {[string length $flt] > 0} {
         if {!$printedText} {
            puts "$portText"
            set printedText 1
         }
         puts "$flt"
      }
   }
}

# Process everything from the configuration file
# (i.e. that should be there)
foreach c $configured {
   set i [snmpIfIndex $c]
   if {$i > 0} {
      processOne $i $c [getPortName $c]
   } else {
      puts "$c:\t[getPortName $c] doesn't exist on this switch"
   }
}

# Now what's actually there, skipping things that
# we've already noted from the configuration file
foreach a $actual {
   if {! [info exists upDown($a)]} {
      if {$doShowDown > 0 || [snmpIfUp $a]} {
         processOne $a "Port $a" "*unknown peer*"
      }
   }
}

# Now do any updates necessary
if {$debug || [llength $genericUpdates] > 0} {
   puts "Generic updates required:"
   foreach u $genericUpdates {
      puts "\t$u"
   }
}
if {$debug || [llength $portUpdates] > 0} {
   puts "Port updates required:"
   foreach u $portUpdates {
      puts "\t$u"
   }
}
if {$debug || [llength $vlanDeletes] > 0} {
   puts "VLAN deletes required:"
   foreach u $vlanDeletes {
      puts "\t$u"
   }
}

set errors 0
if {$doUpdate} {
   foreach u $genericUpdates {
      if {![eval $u]} { incr errors }
      after 1500
   }
   foreach u $portUpdates {
      if {![eval $u]} { incr errors }
      after 1500
   }
   if {$trunkingInUse} {
      puts stderr "Not deleting VLANs as trunking is in use"
      incr errors
   } else {
      foreach u $vlanDeletes {
         if {![eval $u]} { incr errors }
         after 1500
      }
   }
}
if {$errors > 0} {
   puts "Errors: $errors"
}

exit 0
